@@ -21,107 +21,108 @@ import numpy as np
 from PIL import Image
 import tornado
 
-from matplotlib import backend_bases, cbook
+from matplotlib import _api, backend_bases
 from matplotlib.backends import backend_agg
 from matplotlib.backend_bases import _Backend
 
 _log = logging.getLogger(__name__)
 
-# http://www.cambiaresearch.com/articles/15/javascript-char-codes-key-codes
-_SHIFT_LUT = {59: ':',
-              61: '+',
-              173: '_',
-              186: ':',
-              187: '+',
-              188: '<',
-              189: '_',
-              190: '>',
-              191: '?',
-              192: '~',
-              219: '{',
-              220: '|',
-              221: '}',
-              222: '"'}
-
-_LUT = {8: 'backspace',
-        9: 'tab',
-        13: 'enter',
-        16: 'shift',
-        17: 'control',
-        18: 'alt',
-        19: 'pause',
-        20: 'caps',
-        27: 'escape',
-        32: ' ',
-        33: 'pageup',
-        34: 'pagedown',
-        35: 'end',
-        36: 'home',
-        37: 'left',
-        38: 'up',
-        39: 'right',
-        40: 'down',
-        45: 'insert',
-        46: 'delete',
-        91: 'super',
-        92: 'super',
-        93: 'select',
-        106: '*',
-        107: '+',
-        109: '-',
-        110: '.',
-        111: '/',
-        144: 'num_lock',
-        145: 'scroll_lock',
-        186: ':',
-        187: '=',
-        188: ',',
-        189: '-',
-        190: '.',
-        191: '/',
-        192: '`',
-        219: '[',
-        220: '\\',
-        221: ']',
-        222: "'"}
+_SPECIAL_KEYS_LUT = {'Alt': 'alt',
+                     'AltGraph': 'alt',
+                     'CapsLock': 'caps_lock',
+                     'Control': 'control',
+                     'Meta': 'meta',
+                     'NumLock': 'num_lock',
+                     'ScrollLock': 'scroll_lock',
+                     'Shift': 'shift',
+                     'Super': 'super',
+                     'Enter': 'enter',
+                     'Tab': 'tab',
+                     'ArrowDown': 'down',
+                     'ArrowLeft': 'left',
+                     'ArrowRight': 'right',
+                     'ArrowUp': 'up',
+                     'End': 'end',
+                     'Home': 'home',
+                     'PageDown': 'pagedown',
+                     'PageUp': 'pageup',
+                     'Backspace': 'backspace',
+                     'Delete': 'delete',
+                     'Insert': 'insert',
+                     'Escape': 'escape',
+                     'Pause': 'pause',
+                     'Select': 'select',
+                     'Dead': 'dead',
+                     'F1': 'f1',
+                     'F2': 'f2',
+                     'F3': 'f3',
+                     'F4': 'f4',
+                     'F5': 'f5',
+                     'F6': 'f6',
+                     'F7': 'f7',
+                     'F8': 'f8',
+                     'F9': 'f9',
+                     'F10': 'f10',
+                     'F11': 'f11',
+                     'F12': 'f12'}
 
 
 def _handle_key(key):
-    """Handle key codes"""
-    code = int(key[key.index('k') + 1:])
-    value = chr(code)
-    # letter keys
-    if 65 <= code <= 90:
-        if 'shift+' in key:
+    """Handle key values"""
+    value = key[key.index('k') + 1:]
+    if 'shift+' in key:
+        if len(value) == 1:
             key = key.replace('shift+', '')
-        else:
-            value = value.lower()
-    # number keys
-    elif 48 <= code <= 57:
-        if 'shift+' in key:
-            value = ')!@#$%^&*('[int(value)]
-            key = key.replace('shift+', '')
-    # function keys
-    elif 112 <= code <= 123:
-        value = 'f%s' % (code - 111)
-    # number pad keys
-    elif 96 <= code <= 105:
-        value = '%s' % (code - 96)
-    # keys with shift alternatives
-    elif code in _SHIFT_LUT and 'shift+' in key:
-        key = key.replace('shift+', '')
-        value = _SHIFT_LUT[code]
-    elif code in _LUT:
-        value = _LUT[code]
+    if value in _SPECIAL_KEYS_LUT:
+        value = _SPECIAL_KEYS_LUT[value]
     key = key[:key.index('k')] + value
     return key
 
 
+class TimerTornado(backend_bases.TimerBase):
+    def __init__(self, *args, **kwargs):
+        self._timer = None
+        super().__init__(*args, **kwargs)
+
+    def _timer_start(self):
+        self._timer_stop()
+        if self._single:
+            ioloop = tornado.ioloop.IOLoop.instance()
+            self._timer = ioloop.add_timeout(
+                datetime.timedelta(milliseconds=self.interval),
+                self._on_timer)
+        else:
+            self._timer = tornado.ioloop.PeriodicCallback(
+                self._on_timer,
+                max(self.interval, 1e-6))
+            self._timer.start()
+
+    def _timer_stop(self):
+        if self._timer is None:
+            return
+        elif self._single:
+            ioloop = tornado.ioloop.IOLoop.instance()
+            ioloop.remove_timeout(self._timer)
+        else:
+            self._timer.stop()
+        self._timer = None
+
+    def _timer_set_interval(self):
+        # Only stop and restart it if the timer has already been started
+        if self._timer is not None:
+            self._timer_stop()
+            self._timer_start()
+
+
 class FigureCanvasWebAggCore(backend_agg.FigureCanvasAgg):
+    _timer_cls = TimerTornado
+    # Webagg and friends having the right methods, but still
+    # having bugs in practice.  Do not advertise that it works until
+    # we can debug this.
     supports_blit = False
 
     def __init__(self, *args, **kwargs):
-        backend_agg.FigureCanvasAgg.__init__(self, *args, **kwargs)
+        super().__init__(*args, **kwargs)
 
         # Set to True when the renderer contains data that is newer
         # than the PNG buffer.
@@ -153,6 +154,10 @@ class FigureCanvasWebAggCore(backend_agg.FigureCanvasAgg):
         finally:
             self.manager.refresh_all()  # Swap the frames.
 
+    def blit(self, bbox=None):
+        self._png_is_old = True
+        self.manager.refresh_all()
+
     def draw_idle(self):
         self.send_event("draw")
 
@@ -165,7 +170,7 @@ class FigureCanvasWebAggCore(backend_agg.FigureCanvasAgg):
         draw this mode may be changed if the resulting image has any
         transparent component.
         """
-        cbook._check_in_list(['full', 'diff'], mode=mode)
+        _api.check_in_list(['full', 'diff'], mode=mode)
         if self._current_image_mode != mode:
             self._current_image_mode = mode
             self.handle_send_image_mode(None)
@@ -189,21 +194,18 @@ class FigureCanvasWebAggCore(backend_agg.FigureCanvasAgg):
                 output = buff
             else:
                 self.set_image_mode('diff')
-                last_buffer = (np.frombuffer(self._last_renderer.buffer_rgba(),
-                                             dtype=np.uint32)
-                               .reshape((renderer.height, renderer.width)))
-                diff = buff != last_buffer
+                diff = buff != self._last_buff
                 output = np.where(diff, buff, 0)
 
-            buf = BytesIO()
-            data = output.view(dtype=np.uint8).reshape((*output.shape, 4))
-            Image.fromarray(data).save(buf, format="png")
-            # Swap the renderer frames
-            self._renderer, self._last_renderer = (
-                self._last_renderer, renderer)
+            # Store the current buffer so we can compute the next diff.
+            np.copyto(self._last_buff, buff)
             self._force_full = False
             self._png_is_old = False
-            return buf.getvalue()
+
+            data = output.view(dtype=np.uint8).reshape((*output.shape, 4))
+            with BytesIO() as png:
+                Image.fromarray(data).save(png, format="png")
+                return png.getvalue()
 
     def get_renderer(self, cleared=None):
         # Mirrors super.get_renderer, but caches the old one so that we can do
@@ -220,9 +222,10 @@ class FigureCanvasWebAggCore(backend_agg.FigureCanvasAgg):
         if need_new_renderer:
             self._renderer = backend_agg.RendererAgg(
                 w, h, self.figure.dpi)
-            self._last_renderer = backend_agg.RendererAgg(
-                w, h, self.figure.dpi)
             self._lastKey = key
+            self._last_buff = np.copy(np.frombuffer(
+                self._renderer.buffer_rgba(), dtype=np.uint32
+            ).reshape((self._renderer.height, self._renderer.width)))
 
         elif cleared:
             self._renderer.clear()
@@ -260,18 +263,13 @@ class FigureCanvasWebAggCore(backend_agg.FigureCanvasAgg):
         # off by 1
         button = event['button'] + 1
 
-        # The right mouse button pops up a context menu, which
-        # doesn't work very well, so use the middle mouse button
-        # instead.  It doesn't seem that it's possible to disable
-        # the context menu in recent versions of Chrome.  If this
-        # is resolved, please also adjust the docstring in MouseEvent.
-        if button == 2:
-            button = 3
-
         e_type = event['type']
         guiEvent = event.get('guiEvent', None)
         if e_type == 'button_press':
             self.button_press_event(x, y, button, guiEvent=guiEvent)
+        elif e_type == 'dblclick':
+            self.button_press_event(x, y, button, dblclick=True,
+                                    guiEvent=guiEvent)
         elif e_type == 'button_release':
             self.button_release_event(x, y, button, guiEvent=guiEvent)
         elif e_type == 'motion_notify':
@@ -282,9 +280,9 @@ class FigureCanvasWebAggCore(backend_agg.FigureCanvasAgg):
             self.leave_notify_event()
         elif e_type == 'scroll':
             self.scroll_event(x, y, event['step'], guiEvent=guiEvent)
-    handle_button_press = handle_button_release = handle_motion_notify = \
-        handle_figure_enter = handle_figure_leave = handle_scroll = \
-        _handle_mouse
+    handle_button_press = handle_button_release = handle_dblclick = \
+        handle_figure_enter = handle_figure_leave = handle_motion_notify = \
+        handle_scroll = _handle_mouse
 
     def _handle_key(self, event):
         key = _handle_key(event['key'])
@@ -387,7 +385,7 @@ class NavigationToolbar2WebAgg(backend_bases.NavigationToolbar2):
             "rubberband", x0=x0, y0=y0, x1=x1, y1=y1)
 
     def release_zoom(self, event):
-        backend_bases.NavigationToolbar2.release_zoom(self, event)
+        super().release_zoom(event)
         self.canvas.send_event(
             "rubberband", x0=-1, y0=-1, x1=-1, y1=-1)
 
@@ -414,10 +412,8 @@ class FigureManagerWebAgg(backend_bases.FigureManagerBase):
     ToolbarCls = NavigationToolbar2WebAgg
 
     def __init__(self, canvas, num):
-        backend_bases.FigureManagerBase.__init__(self, canvas, num)
-
         self.web_sockets = set()
-
+        super().__init__(canvas, num)
         self.toolbar = self._get_toolbar(canvas)
 
     def show(self):
@@ -481,8 +477,7 @@ class FigureManagerWebAgg(backend_bases.FigureManagerBase):
         for filetype, ext in sorted(FigureCanvasWebAggCore.
                                     get_supported_filetypes_grouped().
                                     items()):
-            if ext[0] != 'pgf':  # pgf does not support BytesIO
-                extensions.append(ext[0])
+            extensions.append(ext[0])
         output.write("mpl.extensions = {0};\n\n".format(
             json.dumps(extensions)))
 
@@ -502,41 +497,6 @@ class FigureManagerWebAgg(backend_bases.FigureManagerBase):
             s.send_json(payload)
 
 
-class TimerTornado(backend_bases.TimerBase):
-    def __init__(self, *args, **kwargs):
-        self._timer = None
-        super().__init__(*args, **kwargs)
-
-    def _timer_start(self):
-        self._timer_stop()
-        if self._single:
-            ioloop = tornado.ioloop.IOLoop.instance()
-            self._timer = ioloop.add_timeout(
-                datetime.timedelta(milliseconds=self.interval),
-                self._on_timer)
-        else:
-            self._timer = tornado.ioloop.PeriodicCallback(
-                self._on_timer,
-                max(self.interval, 1e-6))
-            self._timer.start()
-
-    def _timer_stop(self):
-        if self._timer is None:
-            return
-        elif self._single:
-            ioloop = tornado.ioloop.IOLoop.instance()
-            ioloop.remove_timeout(self._timer)
-        else:
-            self._timer.stop()
-        self._timer = None
-
-    def _timer_set_interval(self):
-        # Only stop and restart it if the timer has already been started
-        if self._timer is not None:
-            self._timer_stop()
-            self._timer_start()
-
-
 @_Backend.export
 class _BackendWebAggCoreAgg(_Backend):
     FigureCanvas = FigureCanvasWebAggCore
