@@ -1,25 +1,30 @@
 """
-Provides a collection of utilities for comparing (image) results.
-
+Utilities for comparing image results.
 """
 
 import atexit
+import functools
 import hashlib
+import logging
 import os
 from pathlib import Path
 import re
 import shutil
 import subprocess
 import sys
-from tempfile import TemporaryFile
+from tempfile import TemporaryDirectory, TemporaryFile
+import weakref
 
 import numpy as np
+from PIL import Image
 
 import matplotlib as mpl
+from matplotlib import _api, cbook
 from matplotlib.testing.exceptions import ImageComparisonFailure
-from matplotlib import cbook
 
-__all__ = ['compare_images', 'comparable_formats']
+_log = logging.getLogger(__name__)
+
+__all__ = ['calculate_rms', 'comparable_formats', 'compare_images']
 
 
 def make_test_filename(fname, purpose):
@@ -30,20 +35,16 @@ def make_test_filename(fname, purpose):
     return '%s-%s%s' % (base, purpose, ext)
 
 
-def get_cache_dir():
-    cachedir = mpl.get_cachedir()
-    if cachedir is None:
-        raise RuntimeError('Could not find a suitable configuration directory')
-    cache_dir = os.path.join(cachedir, 'test_cache')
-    try:
-        Path(cache_dir).mkdir(parents=True, exist_ok=True)
-    except IOError:
-        return None
-    if not os.access(cache_dir, os.W_OK):
-        return None
+def _get_cache_path():
+    cache_dir = Path(mpl.get_cachedir(), 'test_cache')
+    cache_dir.mkdir(parents=True, exist_ok=True)
     return cache_dir
 
 
+def get_cache_dir():
+    return str(_get_cache_path())
+
+
 def get_file_hash(path, block_size=2 ** 20):
     md5 = hashlib.md5()
     with open(path, 'rb') as fd:
@@ -53,16 +54,17 @@ def get_file_hash(path, block_size=2 ** 20):
                 break
             md5.update(data)
 
-    if path.endswith('.pdf'):
+    if Path(path).suffix == '.pdf':
         md5.update(str(mpl._get_executable_info("gs").version)
                    .encode('utf-8'))
-    elif path.endswith('.svg'):
+    elif Path(path).suffix == '.svg':
         md5.update(str(mpl._get_executable_info("inkscape").version)
                    .encode('utf-8'))
 
     return md5.hexdigest()
 
 
+@_api.deprecated("3.3")
 def make_external_conversion_command(cmd):
     def convert(old, new):
         cmdline = cmd(old, new)
@@ -122,7 +124,7 @@ class _Converter:
                 raise _ConverterError
             buf.extend(c)
             if buf.endswith(terminator):
-                return bytes(buf[:-len(terminator)])
+                return bytes(buf)
 
 
 class _GSConverter(_Converter):
@@ -135,8 +137,8 @@ class _GSConverter(_Converter):
                 stdin=subprocess.PIPE, stdout=subprocess.PIPE)
             try:
                 self._read_until(b"\nGS")
-            except _ConverterError:
-                raise OSError("Failed to start Ghostscript")
+            except _ConverterError as err:
+                raise OSError("Failed to start Ghostscript") from err
 
         def encode_and_escape(name):
             return (os.fsencode(name)
@@ -152,69 +154,88 @@ class _GSConverter(_Converter):
             + b") run flush\n")
         self._proc.stdin.flush()
         # GS> if nothing left on the stack; GS<n> if n items left on the stack.
-        err = self._read_until(b"GS")
-        stack = self._read_until(b">")
+        err = self._read_until((b"GS<", b"GS>"))
+        stack = ""
+        if err.endswith(b"GS<"):
+            stack = self._read_until(b">")
         if stack or not os.path.exists(dest):
-            stack_size = int(stack[1:]) if stack else 0
+            stack_size = int(stack[:-1]) if stack else 0
             self._proc.stdin.write(b"pop\n" * stack_size)
             # Using the systemencoding should at least get the filenames right.
             raise ImageComparisonFailure(
-                (err + b"GS" + stack + b">")
-                .decode(sys.getfilesystemencoding(), "replace"))
+                (err + stack).decode(sys.getfilesystemencoding(), "replace"))
 
 
 class _SVGConverter(_Converter):
     def __call__(self, orig, dest):
+        old_inkscape = mpl._get_executable_info("inkscape").version < "1"
+        terminator = b"\n>" if old_inkscape else b"> "
+        if not hasattr(self, "_tmpdir"):
+            self._tmpdir = TemporaryDirectory()
+            # On Windows, we must make sure that self._proc has terminated
+            # (which __del__ does) before clearing _tmpdir.
+            weakref.finalize(self._tmpdir, self.__del__)
         if (not self._proc  # First run.
                 or self._proc.poll() is not None):  # Inkscape terminated.
-            env = os.environ.copy()
-            # If one passes e.g. a png file to Inkscape, it will try to
-            # query the user for conversion options via a GUI (even with
-            # `--without-gui`).  Unsetting `DISPLAY` prevents this (and causes
-            # GTK to crash and Inkscape to terminate, but that'll just be
-            # reported as a regular exception below).
-            env.pop("DISPLAY", None)  # May already be unset.
-            # Do not load any user options.
-            env["INKSCAPE_PROFILE_DIR"] = os.devnull
-            # Old versions of Inkscape (0.48.3.1, used on Travis as of now)
-            # seem to sometimes deadlock when stderr is redirected to a pipe,
-            # so we redirect it to a temporary file instead.  This is not
-            # necessary anymore as of Inkscape 0.92.1.
+            env = {
+                **os.environ,
+                # If one passes e.g. a png file to Inkscape, it will try to
+                # query the user for conversion options via a GUI (even with
+                # `--without-gui`).  Unsetting `DISPLAY` prevents this (and
+                # causes GTK to crash and Inkscape to terminate, but that'll
+                # just be reported as a regular exception below).
+                "DISPLAY": "",
+                # Do not load any user options.
+                "INKSCAPE_PROFILE_DIR": os.devnull,
+            }
+            # Old versions of Inkscape (e.g. 0.48.3.1) seem to sometimes
+            # deadlock when stderr is redirected to a pipe, so we redirect it
+            # to a temporary file instead.  This is not necessary anymore as of
+            # Inkscape 0.92.1.
             stderr = TemporaryFile()
             self._proc = subprocess.Popen(
-                ["inkscape", "--without-gui", "--shell"],
-                stdin=subprocess.PIPE, stdout=subprocess.PIPE,
-                stderr=stderr, env=env)
+                ["inkscape", "--without-gui", "--shell"] if old_inkscape else
+                ["inkscape", "--shell"],
+                stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=stderr,
+                env=env, cwd=self._tmpdir.name)
             # Slight abuse, but makes shutdown handling easier.
             self._proc.stderr = stderr
             try:
-                self._read_until(b"\n>")
-            except _ConverterError:
-                raise OSError("Failed to start Inkscape in interactive mode")
+                self._read_until(terminator)
+            except _ConverterError as err:
+                raise OSError("Failed to start Inkscape in interactive "
+                              "mode") from err
 
-        # Inkscape uses glib's `g_shell_parse_argv`, which has a consistent
-        # behavior across platforms, so we can just use `shlex.quote`.
-        orig_b, dest_b = map(_shlex_quote_bytes,
-                             map(os.fsencode, [orig, dest]))
-        if b"\n" in orig_b or b"\n" in dest_b:
-            # Who knows whether the current folder name has a newline, or if
-            # our encoding is even ASCII compatible...  Just fall back on the
-            # slow solution (Inkscape uses `fgets` so it will always stop at a
-            # newline).
-            return make_external_conversion_command(lambda old, new: [
-                'inkscape', '-z', old, '--export-png', new])(orig, dest)
-        self._proc.stdin.write(orig_b + b" --export-png=" + dest_b + b"\n")
+        # Inkscape's shell mode does not support escaping metacharacters in the
+        # filename ("\n", and ":;" for inkscape>=1).  Avoid any problems by
+        # running from a temporary directory and using fixed filenames.
+        inkscape_orig = Path(self._tmpdir.name, os.fsdecode(b"f.svg"))
+        inkscape_dest = Path(self._tmpdir.name, os.fsdecode(b"f.png"))
+        try:
+            inkscape_orig.symlink_to(Path(orig).resolve())
+        except OSError:
+            shutil.copyfile(orig, inkscape_orig)
+        self._proc.stdin.write(
+            b"f.svg --export-png=f.png\n" if old_inkscape else
+            b"file-open:f.svg;export-filename:f.png;export-do;file-close\n")
         self._proc.stdin.flush()
         try:
-            self._read_until(b"\n>")
-        except _ConverterError:
+            self._read_until(terminator)
+        except _ConverterError as err:
             # Inkscape's output is not localized but gtk's is, so the output
             # stream probably has a mixed encoding.  Using the filesystem
             # encoding should at least get the filenames right...
-            self._stderr.seek(0)
+            self._proc.stderr.seek(0)
             raise ImageComparisonFailure(
-                self._stderr.read().decode(
-                    sys.getfilesystemencoding(), "replace"))
+                self._proc.stderr.read().decode(
+                    sys.getfilesystemencoding(), "replace")) from err
+        os.remove(inkscape_orig)
+        shutil.move(inkscape_dest, dest)
+
+    def __del__(self):
+        super().__del__()
+        if hasattr(self, "_tmpdir"):
+            self._tmpdir.cleanup()
 
 
 def _update_converter():
@@ -247,7 +268,7 @@ def comparable_formats():
 
     Returns
     -------
-    supported_formats : list of str
+    list of str
         E.g. ``['png', 'pdf', 'svg', 'eps']``.
 
     """
@@ -260,40 +281,68 @@ def convert(filename, cache):
 
     If *cache* is True, the result of the conversion is cached in
     `matplotlib.get_cachedir() + '/test_cache/'`.  The caching is based on a
-    hash of the exact contents of the input file.  There is no limit on the
-    size of the cache, so it may need to be manually cleared periodically.
+    hash of the exact contents of the input file.  Old cache entries are
+    automatically deleted as needed to keep the size of the cache capped to
+    twice the size of all baseline images.
     """
-    base, extension = os.fspath(filename).rsplit('.', 1)
-    if extension not in converter:
+    path = Path(filename)
+    if not path.exists():
+        raise IOError(f"{path} does not exist")
+    if path.suffix[1:] not in converter:
         import pytest
-        pytest.skip(f"Don't know how to convert {extension} files to png")
-    newname = base + '_' + extension + '.png'
-    if not os.path.exists(filename):
-        raise IOError("'%s' does not exist" % filename)
+        pytest.skip(f"Don't know how to convert {path.suffix} files to png")
+    newpath = path.parent / f"{path.stem}_{path.suffix[1:]}.png"
 
     # Only convert the file if the destination doesn't already exist or
     # is out of date.
-    if (not os.path.exists(newname) or
-            os.stat(newname).st_mtime < os.stat(filename).st_mtime):
-        if cache:
-            cache_dir = get_cache_dir()
-        else:
-            cache_dir = None
+    if not newpath.exists() or newpath.stat().st_mtime < path.stat().st_mtime:
+        cache_dir = _get_cache_path() if cache else None
 
         if cache_dir is not None:
-            hash_value = get_file_hash(filename)
-            new_ext = os.path.splitext(newname)[1]
-            cached_file = os.path.join(cache_dir, hash_value + new_ext)
-            if os.path.exists(cached_file):
-                shutil.copyfile(cached_file, newname)
-                return newname
+            _register_conversion_cache_cleaner_once()
+            hash_value = get_file_hash(path)
+            cached_path = cache_dir / (hash_value + newpath.suffix)
+            if cached_path.exists():
+                _log.debug("For %s: reusing cached conversion.", filename)
+                shutil.copyfile(cached_path, newpath)
+                return str(newpath)
 
-        converter[extension](filename, newname)
+        _log.debug("For %s: converting to png.", filename)
+        converter[path.suffix[1:]](path, newpath)
 
         if cache_dir is not None:
-            shutil.copyfile(newname, cached_file)
+            _log.debug("For %s: caching conversion result.", filename)
+            shutil.copyfile(newpath, cached_path)
 
-    return newname
+    return str(newpath)
+
+
+def _clean_conversion_cache():
+    # This will actually ignore mpl_toolkits baseline images, but they're
+    # relatively small.
+    baseline_images_size = sum(
+        path.stat().st_size
+        for path in Path(mpl.__file__).parent.glob("**/baseline_images/**/*"))
+    # 2x: one full copy of baselines, and one full copy of test results
+    # (actually an overestimate: we don't convert png baselines and results).
+    max_cache_size = 2 * baseline_images_size
+    # Reduce cache until it fits.
+    with cbook._lock_path(_get_cache_path()):
+        cache_stat = {
+            path: path.stat() for path in _get_cache_path().glob("*")}
+        cache_size = sum(stat.st_size for stat in cache_stat.values())
+        paths_by_atime = sorted(  # Oldest at the end.
+            cache_stat, key=lambda path: cache_stat[path].st_atime,
+            reverse=True)
+        while cache_size > max_cache_size:
+            path = paths_by_atime.pop()
+            cache_size -= cache_stat[path].st_size
+            path.unlink()
+
+
+@functools.lru_cache()  # Ensure this is only registered once.
+def _register_conversion_cache_cleaner_once():
+    atexit.register(_clean_conversion_cache)
 
 
 def crop_to_same(actual_path, actual_image, expected_path, expected_image):
@@ -308,7 +357,9 @@ def crop_to_same(actual_path, actual_image, expected_path, expected_image):
 
 
 def calculate_rms(expected_image, actual_image):
-    "Calculate the per-pixel errors, then compute the root mean square error."
+    """
+    Calculate the per-pixel errors, then compute the root mean square error.
+    """
     if expected_image.shape != actual_image.shape:
         raise ImageComparisonFailure(
             "Image sizes do not match expected size: {} "
@@ -317,6 +368,10 @@ def calculate_rms(expected_image, actual_image):
     return np.sqrt(((expected_image - actual_image).astype(float) ** 2).mean())
 
 
+# NOTE: compare_image and save_diff_image assume that the image does not have
+# 16-bit depth, as Pillow converts these to RGB incorrectly.
+
+
 def compare_images(expected, actual, tol, in_decorator=False):
     """
     Compare two "image" files checking differences within a tolerance.
@@ -341,7 +396,7 @@ def compare_images(expected, actual, tol, in_decorator=False):
 
     Returns
     -------
-    comparison_result : None or dict or str
+    None or dict or str
         Return *None* if the images are equal within the given tolerance.
 
         If the images differ, the return value depends on  *in_decorator*.
@@ -366,8 +421,6 @@ def compare_images(expected, actual, tol, in_decorator=False):
         compare_images(img1, img2, 0.001)
 
     """
-    from matplotlib import _png
-
     actual = os.fspath(actual)
     if not os.path.exists(actual):
         raise Exception("Output image %s does not exist." % actual)
@@ -380,14 +433,12 @@ def compare_images(expected, actual, tol, in_decorator=False):
         raise IOError('Baseline image %r does not exist.' % expected)
     extension = expected.split('.')[-1]
     if extension != 'png':
-        actual = convert(actual, False)
-        expected = convert(expected, True)
+        actual = convert(actual, cache=True)
+        expected = convert(expected, cache=True)
 
     # open the image files and remove the alpha channel (if it exists)
-    with open(expected, "rb") as expected_file:
-        expected_image = _png.read_png_int(expected_file)[:, :, :3]
-    with open(actual, "rb") as actual_file:
-        actual_image = _png.read_png_int(actual_file)[:, :, :3]
+    expected_image = np.asarray(Image.open(expected).convert("RGB"))
+    actual_image = np.asarray(Image.open(actual).convert("RGB"))
 
     actual_image, expected_image = crop_to_same(
         actual, actual_image, expected, expected_image)
@@ -426,7 +477,7 @@ def compare_images(expected, actual, tol, in_decorator=False):
 
 
 def save_diff_image(expected, actual, output):
-    '''
+    """
     Parameters
     ----------
     expected : str
@@ -435,13 +486,10 @@ def save_diff_image(expected, actual, output):
         File path of actual image.
     output : str
         File path to save difference image to.
-    '''
+    """
     # Drop alpha channels, similarly to compare_images.
-    from matplotlib import _png
-    with open(expected, "rb") as expected_file:
-        expected_image = _png.read_png(expected_file)[..., :3]
-    with open(actual, "rb") as actual_file:
-        actual_image = _png.read_png(actual_file)[..., :3]
+    expected_image = np.asarray(Image.open(expected).convert("RGB"))
+    actual_image = np.asarray(Image.open(actual).convert("RGB"))
     actual_image, expected_image = crop_to_same(
         actual, actual_image, expected, expected_image)
     expected_image = np.array(expected_image).astype(float)
@@ -467,5 +515,4 @@ def save_diff_image(expected, actual, output):
     # Hard-code the alpha channel to fully solid
     save_image_np[:, :, 3] = 255
 
-    with open(output, "wb") as output_file:
-        _png.write_png(save_image_np, output_file)
+    Image.fromarray(save_image_np).save(output, format="png")
